#!/bin/sh

# The bot, doing all the update and sync operations
# (c) 2008-2010, Fabien Tassin <fta@sofaraway.org>

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2, or (at
# your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

set -e
set -x
set -u

WORKDIR="$(readlink -f $(dirname $0)/../work)"
BINDIR="$(readlink -f $(dirname $0))"
CONFDIR=$WORKDIR/confs
UPGRADE_OPTS=""
SYNC_OPTS=""
START_WD=$(pwd)

#######

get_conf () {
  local conf=$1
  local key=$2
  perl -e '
    $/ = undef;
    open(F, "'$conf'") || print "Cant open conf file: $!\n";
    $s = <F>;
    close F;
    $data = eval $s;
    my @keys = split /\//, "'$key'";
    $d = $data;
    while (my $k = shift @keys) {
      exit 1 unless defined $$d{$k};
      $d = $$d{$k};
      if (ref($d) eq "ARRAY" && $#$keys == -1) {
        $d = join " ", @$d;
        last;
      }
    }
    print $d;
    exit 0;
  '
  return 0
}

send_email () {
  return
  local dir=$1
  local pocket=$2
  local from="`echo $3 | sed -e 's/@/\\\\@/g'`"
  local to="`echo $4 | sed -e 's/@/\\\\@/g'`"
  # Prepare a Summary based on the log files
  echo "Summary for '$pocket'\n" > $dir/summary.txt
  UPLOAD=0
  STATUS=OK
  if [ -s $dir/$pocket-update.txt ] ; then
    C=0
    if [ `grep -c '\*LOG\* New upstream snapshot' $dir/$pocket-update.txt` -gt 0 ] ; then
      echo '* new upstream snapshots:' >> $dir/summary.txt
      grep '\*LOG\* New upstream snapshot' $dir/$pocket-update.txt | cut -d' ' -f5- | sed -e 's/^/  - /' >> $dir/summary.txt
      C=1
    fi
    if [ `grep -c 'No new upstream snapshot' $dir/$pocket-update.txt` -gt 0 ] ; then
      echo '* no new upstream snapshots:' >> $dir/summary.txt
      grep 'No new upstream snapshot' $dir/$pocket-update.txt | cut -d' ' -f6- | sed -e 's/^/  - /' >> $dir/summary.txt
      C=1
    fi
    if [ $C -ne 0 ] ; then
      echo >> $dir/summary.txt
    fi
    C=0
    if [ `grep -c '\*LOG\* update' $dir/$pocket-update.txt` -gt 0 ] ; then
      echo '* New packaging revisions:' >> $dir/summary.txt
      grep '\*LOG\* update' $dir/$pocket-update.txt | cut -d' ' -f3- | sed -e 's/^/  - /' |
        sed -e 's/^  - [^ @]*@revs: /    \* /' >> $dir/summary.txt
      C=1
    fi
    if [ `grep -Ec '\**ERROR* (src package differs|version)' $dir/$pocket-update.txt` -gt 0 ] ; then
      echo '* New upstream snapshots errors:' >> $dir/summary.txt
      grep -E '\**ERROR* (src package differs|version)' $dir/$pocket-update.txt | cut -d' ' -f2- | sed -e 's/^/  - /' >> $dir/summary.txt
      C=1
    fi
    if [ $C -ne 0 ] ; then
      echo >> $dir/summary.txt
    fi
  else
    echo "*ERROR* no $pocket-update.txt log? please investigate\n" >> $dir/summary.txt
  fi
  if [ -s "$dir/$pocket-sync.txt" ] ; then
    C=0
    if [ `grep -c '\*LOG\* sync'  $dir/$pocket-sync.txt` -gt 0 ] ; then
      echo '* sync results:' >> $dir/summary.txt
      grep '\*LOG\* sync' $dir/$pocket-sync.txt | cut -d' ' -f3-  | sed -e 's/^/  - /' >> $dir/summary.txt
      C=1
    fi
    if [ $C -ne 0 ] ; then
      echo >> $dir/summary.txt
    fi
    C=0
    if [ `grep -c '\*LOG\* dput' $dir/$pocket-sync.txt` -gt 0 ] ; then
      echo '* dput results:' >> $dir/summary.txt
      grep '\*LOG\* dput' $dir/$pocket-sync.txt | cut -d' ' -f2-  | sed -e 's/^/  - /' >> $dir/summary.txt
      if [ `grep -c 'dput successfully pushed' $dir/summary.txt` -gt 0 ] ; then
        UPLOAD=1
      fi
      C=1
    fi
    if [ $C -ne 0 ] ; then
      echo >> $dir/summary.txt
    fi
  elif [ $no_sync -ne 1 ] ; then
    echo "*ERROR* no $pocket-sync.txt log? please investigate\n" >> $dir/summary.txt
  fi
  grep -qE '(\*ERROR\*|Abort|returned with error|check for errors)' $dir/summary.txt
  if [ $? -eq 0 ] ; then
    STATUS=FAILED
  fi
  if [ $UPLOAD -eq 0 ] ; then
    STATUS=$STATUS-NOUPLOAD
  fi
  for file in $dir/*.txt ; do
    if [ `stat -c %s $file` -gt 100000 ] ; then
      gzip -9 $file
    fi
  done
  perl -e '
    use MIME::Entity;
    my $dir = "'"$dir"'";
    my $status = "'"$STATUS"'";
    my $top = MIME::Entity->build(Type    => "multipart/mixed",
                                  From    => "'"$from"'",
                                  To      => "'"$to"'",
                                  Subject => "[PPABOT] Logs for '$pocket' ($status)",
                                  "X-ppabot" => "pocket='$pocket'");

    opendir(DIR, "$dir") || die "Cant opendir $dir: $!";
    my @files = sort { -M "$dir/$b" <=> -M "$dir/$a" } readdir(DIR);
    closedir DIR;
    open(F, "$dir/summary.txt");
    local $/ = undef;
    my $summary = <F>;
    close F;
    $top->attach(Data => $summary);
    for $file (@files) {
      next unless -f "$dir/$file";
      next if -z "$dir/$file";
      next if $file eq "summary.txt";
      if ($file =~ m/\.gz$/) {
        $top->attach(Path     => "$dir/$file",
                     Type     => "application/x-gzip",
                     Encoding => "base64");
      }
      else {
        $top->attach(Path     => "$dir/$file");
      }
    }
    open MAIL, "| /usr/lib/sendmail -t -oi -oem" or die "Cant open sendmail: $!";
    $top->print(\*MAIL);
    close MAIL;
  '
  return 0
}

usage () {
  echo "Usage: "`basename $0`" [-n|--no-tarball-update] [-f|--force] pocket1 [pocket2] [...]"
  echo
  echo "   -e or --email                Email logs to the pocket watchers"
  echo "   -n or --no-tarball-update    Don't update tarballs"
  echo "   -s or --no-sync              Don't sync the branches (implies --no-push)"
  echo "   -p no --no-push              Prevent the final dput to the builders"
  echo "   -f or --force                Force a sync"
  echo "   -h or --help                 This help screen"
}

no_push=0
no_sync=0
no_email=1
while [ $# -gt 0 ]; do
  case "$1" in
    -e | --email )
      shift
      no_email=0 ;;
    -n | --no-tarball-update )
      shift
      UPGRADE_OPTS="$UPGRADE_OPTS -N" ;;
    -s | --no-sync )
      shift
      no_sync=1 ;;
    -p | --no-push )
      shift
      no_push=1 ;;
    -f | --force )
      shift
      SYNC_OPTS="$SYNC_OPTS -f" ;;
    -h | --help )
      usage
      exit 0
      ;;
    -- ) # Stop option processing
      shift
      break ;;
    -* )
      echo Error: unknown option $1
      usage
      exit 1 ;;
    * )
      break ;;
  esac
done

POCKETS="$@"
if [ "Z$POCKETS" = Z ] ; then
  usage
  exit 1
fi

if [ $no_push -eq 0 ] ; then
  SYNC_OPTS="$SYNC_OPTS -P"
fi

cd $WORKDIR

#FOUND=0
#for POCKET in $POCKETS ; do
#  BUILDDIR=`get_conf "$CONFDIR/ppabot-$POCKET.conf" "build_dir"`
#  if [ Z"$BUILDDIR" != Z ] ; then
#    # We don't want to let this dir grow
#    rm -rf $BUILDDIR
#    FOUND=1
#  fi
#done
#if [ $FOUND -eq 0 ] ; then
#  # default build_dir value
#  rm -rf ppa
#fi

for POCKET in $POCKETS ; do
  CONF_OPTS="-p $CONFDIR/ppabot-pkgs-$POCKET.conf -c $CONFDIR/ppabot-$POCKET.conf"
  if [ $no_email -eq 0 ] ; then
    EMAIL_DIR=`mktemp -d`
    LOG_INIT=$EMAIL_DIR/$POCKET-init.txt
    LOG_UPDATE=$EMAIL_DIR/$POCKET-update.txt
    LOG_SYNC=$EMAIL_DIR/$POCKET-sync.txt
    LOG_CLEAN=$EMAIL_DIR/$POCKET-cleanup.txt
  else
    LOG_INIT=/dev/stdout
    LOG_UPDATE=/dev/stdout
    LOG_SYNC=/dev/stdout
    LOG_CLEAN=/dev/stdout
  fi

  # Initialize, in case those are new branches
  $BINDIR/sync-ppa.pl $CONF_OPTS -i 2>&1 |cat -n   #> $LOG_INIT 2>&1

  # Fetch new sources from upstream, if any
  $BINDIR/update-pkg.sh $CONF_OPTS $UPGRADE_OPTS -v > $LOG_UPDATE 2>&1

  if [ $no_sync -eq 0 ] ; then
    # Sync the branches and dput to the configured location (if -P)
    $BINDIR/sync-ppa.pl $CONF_OPTS $SYNC_OPTS 2>&1 |cat -n   #> $LOG_SYNC 2>&1
  fi

  # Clean-up the tarballs directory, preserve only the freshest tarballs
#  PKGS=`get_conf "$CONFDIR/ppabot-pkgs-$POCKET.conf" "order"`
#  TARBALLS=`get_conf "$CONFDIR/ppabot-$POCKET.conf" "tarballs_dir"`
#  if [ "Z$TARBALLS" = Z ]; then
#    TARBALLS=tarballs
#  fi
#  cd $START_WD
#  cd $TARBALLS
#  for managed_package_name in $PKGS; do
#  	for pkg_tarball1 in "${managed_package_name}_"*".orig.tar.gz"; do
#		pt1_ver=$(echo "$pkg_tarball1" |sed -e 's/.*_\(.*\)\.orig.tar.gz$/\1/')
#		for pkg_tarball2 in "${managed_package_name}_"*".orig.tar.gz"; do
#			pt2_ver=$(echo "$pkg_tarball2" |sed -e 's/.*_\(.*\)\.orig.tar.gz$/\1/')
#			if dpkg --compare-versions "$pt1_ver" lt "$pt2_ver"; then
#				rm -vf "$pkg_tarball1"
#				continue 2
#			fi
#		done
#	done
#  done
#  cd -

  # Send an email to all watchers, if we have to
  if [ $no_email -eq 0 ] ; then
    FROM=`get_conf "$CONFDIR/ppabot-$POCKET.conf" "who"`
    WATCHERS=`get_conf "$CONFDIR/ppabot-$POCKET.conf" "watchers"`
    if [ "Z$WATCHERS" = Z ] ; then 
      WATCHERS="$FROM"
    fi
    find $EMAIL_DIR -type f -size 0 -name \*txt -delete
    send_email $EMAIL_DIR $POCKET "$FROM" "$WATCHERS"
    rm -rf $EMAIL_DIR
  fi
done
